@@ -4,8 +4,11 @@ Unit and integration testing in a notebook
 *** Building and running suites of tests ***
 
 class Suite
+    ^^^ Look up its magic function feature!
     Method test (context manager)
 Function fail
+Function assert_
+Predicates same, eq, diff, between, approx, belong_to, is_any_of
 
 *** Reporting test results ***
 
@@ -24,6 +27,7 @@ Class Result
         Sub-class Failure
     Method is_success
     Method is_failure
+Class Explanation
 Class Frame
 
 *** Customizing result reporting ***
@@ -38,25 +42,24 @@ Type TestNameFormatter
     Function ladder
     Function name_all
     Function quoter
-Function detail_results
+Function detail_result
 Function print_frame
 
-*** Checking out the intricacies of test isolation ***
-
-Function protect_environment
-
 Happy jupytesting!
 """
 
 from abc import ABC, abstractmethod
-from copy import copy, deepcopy
+from copy import copy
 from inspect import getframeinfo, Traceback, unwrap
 from io import TextIOBase
 import itertools
 from linecache import getline
+import operator
+import re
 import sys
 from traceback import walk_tb
-from typing import Dict, List, Tuple, Iterator, Union, Iterable, Optional, Any, Callable, Mapping, Sequence
+from typing import Dict, List, Tuple, Iterator, Union, Iterable, Optional, Any, Callable, Mapping, Sequence, cast,\
+        Container
 
 import colors
 from IPython import get_ipython
@@ -113,9 +116,16 @@ class Frame:
     def __init__(self, tb: Traceback, num_line: int, tags: Optional[List[str]] = None) -> None:
         self.num_line = num_line
         self.name_file = tb.filename
-        self.function = tb.function
+        self.function = "" if tb.function == "<module>" else tb.function
         self.tags = tags or []
 
+    def name_file_pretty(self, capitalized: bool = False) -> str:
+        m = re.match(r"<ipython-input-(\d+)-[a-f0-9A-F]+>", self.name_file)
+        if m:
+            name = f"code cell {m.group(1)}"
+            return name.capitalize() if capitalized else name
+        return self.name_file
+
     def context(self, before: int = 3, after: int = 3) -> List[Tuple[int, str]]:
         ctx = [(self.num_line, getline(self.name_file, self.num_line).rstrip())]
         for delta in range(1, before + 1):
@@ -137,7 +147,7 @@ class Frame:
 
     def as_dict(self, context_before: int = 3, context_after: int = 3) -> Dict:
         return {
-            "file": self.name_file,
+            "file": self.name_file_pretty(False),
             "line": self.num_line,
             "function": self.function,
             "context": [[i, line] for i, line in self.context(context_before, context_after)],
@@ -145,7 +155,7 @@ class Frame:
         }
 
     def __str__(self) -> str:
-        return f"File {self.name_file}, Line {self.num_line}, Function {self.function}"
+        return f"{self.function + ': ' if self.function else ''}File {self.name_file}, Line {self.num_line}"
 
     def __repr__(self) -> str:
         return str(self)
@@ -208,36 +218,278 @@ class Error(Result):
         )
         return d
 
+NameArg = Union[int, str]
+Arg = Tuple[NameArg, Any]
+Args = Sequence[Arg]
+
+
+def join_args(args: Sequence[Any], kwargs: Mapping[str, Any]) -> Args:
+    return list(enumerate(args)) + sorted(list(kwargs.items()), key=lambda arg: arg[0])
+
+class Explanation:
+    """
+    Embodies the reason why a predicate fails, including a main expectation that was not satisfied and
+    a list of assertion arguments that tripped the failure.
+    """
+
+    def __init__(self, main: str, culprits: Sequence[Union[Arg, Tuple[str]]]) -> None:
+        self.main = main
+        self.culprits = culprits
+
+    def __repr__(self) -> str:
+        return f"{type(self).__name__}({repr(self.main)}, {repr(self.culprits)})"
+
+    def __bool__(self) -> bool:
+        """
+        All explanations are for failures, so when a predicate returns an explanation, it's as if it had
+        returned False.
+        """
+        return False
+
+    def as_dict(self) -> Dict:
+        return {
+            "main": self.main,
+            "culprits": [[name, str(value)] for name, value in self.culprits]
+        }
+
+class DetailedAssertionError(AssertionError):
+
+    def __init__(self, explanation: Explanation):
+        super().__init__(explanation.main)
+        self.explanation = explanation
+
 class Failure(Error):
     """
     Test result stemming from a condition check that failed, or a test run marked
     as a failure.
     """
-    def __init__(self, reason: str, fns_common: Iterable[Callable]):
+    def __init__(self, explanation: Explanation, fns_common: Iterable[Callable]):
         super().__init__(fns_common)
-        self._reason = reason
+        self._explanation = explanation
+
+    @property
+    def explanation(self) -> Explanation:
+        "Reason given by the programmer as to why the test failed."
+        return self._explanation
 
     @property
     def reason(self) -> str:
-        "Reason given by the programmer as to why the test failed."
-        return self._reason
+        reason = self.explanation.main
+        if self.explanation.culprits:
+            reason += ": " + ", ".join(str(value) for _, value in self.explanation.culprits)
+        return reason
 
     def is_failure(self) -> bool:
         return True
 
     def as_dict(self, context_before: int = 3, context_after: int = 3) -> Dict:
         d = super().as_dict(context_before, context_after)
-        d["reason"] = self.reason
+        d["explanation"] = self.explanation.as_dict()
         return d
 
+ExplanationOnFailure = Union[bool, str, Explanation]
+Predicate = Callable[..., ExplanationOnFailure]
+
+
+def assert_(predicate: Predicate, *ops: Any, msg: str = "", **kwops: Any) -> None:
+    """
+    Evalutes a condition over multiple objects, using as condition test either an explicit function predicate
+    or a bound method provided as first (left-most) operand. Raises an AssertionError exception if the condition
+    is not met (the condition test yields a result of False).
+
+    cond
+        Predicate function (or bound method). It must take as input parameters. The function should return True
+        to signal satisfaction of the predicate. In the other case, it may return a False-equivalent value (in which
+        case the failure explanation is provided as the `msg` keyword argument), or a string corresponding to the
+        explanation, or an ExplanationFailure instance.
+    msg
+        Facultative explanation for the failure of satisfying the predicate. If provided, it will override the
+        explanation returned by the predicate, or at least its "main expectation" (see docstring of
+        `ExplanationFailure`).
+    ops, kwops
+        Operands to the predicate.
+    """
+    result = predicate(*ops, **kwops)
+    if not result:
+        if isinstance(result, Explanation):
+            explanation = result
+        else:
+            explanation = Explanation(result or "", join_args(ops, kwops))
+
+        if msg:
+            explanation.main = msg
+        raise DetailedAssertionError(explanation)
+
+def same(*args, **kwargs) -> ExplanationOnFailure:
+    """
+    Predicate satisfied when all its arguments are the same (according to operator `is`).
+    """
+    return _assess_sequence_2more(
+        "These objects are not the same as {name} => {value}",
+        lambda p, r: None if p is r else hex(hash(p)),
+        args,
+        kwargs
+    )
+
+def _assess_sequence_2more(
+    expl_main: str,
+    pred_repr: Callable[[Any], Any],
+    args: Iterator[Any],
+    kwargs: Mapping[str, Any]
+) -> ExplanationOnFailure:
+    try:
+        (name_ref, value_ref), *rest = join_args(args, kwargs)
+        culprits = [(name_ref, pred_repr(value_ref, None))]
+        for name_comp, value_comp in rest:
+            r = pred_repr(value_comp, value_ref)
+            if r is not None:
+                culprits.append((name_comp, r))
+
+        if len(culprits) > 1:
+            return Explanation(expl_main, culprits)
+        return True
+    except ValueError:
+        raise ValueError("Should provide at least one test argument to this function.")
+
+def eq(*args: Any, **kwargs: Any) -> ExplanationOnFailure:
+    """
+    Exact equality predicate.
+    """
+    return _assess_sequence_2more(
+        "These objects are not equal",
+        lambda p, r: None if p == r else p,
+        args,
+        kwargs
+    )
+
+def diff(*args: Any, **kwargs: Any) -> ExplanationOnFailure:
+    """
+    Difference predicate.
+    """
+    all_args = join_args(args, kwargs)
+    culprits_by_value: List[Tuple[arg, Set[str]]] = []
+    for i, (name_ref, value_ref) in enumerate(all_args[:-1]):
+        for name_comp, value_comp in all_args[i + 1:]:
+            if value_ref == value_comp:
+                try:
+                    names_equal = [s for v, s in culprits_by_value if v == value_ref][0]
+                except IndexError:
+                    culprits_by_value.append((value_ref, set()))
+                    _, names_equal = culprits_by_value[-1]
+                names_equal |= {str(n) for n in [name_ref, name_comp]}
+
+    if len(culprits_by_value) > 0:
+        return Explanation(
+            "Some of the arguments were not mutually different",
+            sorted([(", ".join(sorted(list(names))), value) for value, names in culprits_by_value])
+        )
+    return True
+
+def between(lower: Any, upper: Any, bounds: str = "[]") -> Predicate:
+    """
+    Returns a predicate that is satisfied if all its arguments fall within an interval. Comparands
+    must all support order comparison operators (<, >, <=, >=).
+
+    lower
+        Lower bound of the interval.
+    upper
+        Upper bound of the interval.
+    bounds
+        Determines whether the interval is closed or open at either end (e.g. the boundary element is belongs or not
+        to the set formed by the interval). It is provided as a string whose first and last characters are observed.
+            First character:
+                [    Interval is closed on the left.
+                (    Interval is open on the left.
+            Last character:
+                ]    Interval is closed on the right.
+                )    Interval is open on the right.
+        Any other character in these positions raise an error.
+    """
+    if lower > upper:
+        raise ValueError("Lower bound should be lesser than the upper bound.")
+    cmp_lower, cmp_upper = _get_comparators(bounds)
+    return _assessor_sequence(
+        f"These objects did not belong to interval {bounds[0]}{lower}, {upper}{bounds[1]}",
+        lambda v: cmp_lower(v, lower) and cmp_upper(v, upper),
+    )
+
+
+Comparator = Callable[[Any, Any], bool]
+
+
+def _get_comparators(bounds: str) -> Tuple[Comparator, Comparator]:
+    if not re.match(r"^(\[|\().*(\]|\))$", bounds):
+        raise ValueError("Invalid open/close bounds descriptor")
+    return (
+        {"[": operator.ge, "(": operator.gt}[bounds[0]],
+        {"]": operator.le, ")": operator.lt}[bounds[-1]]
+    )
+
+def _assessor_sequence(expl_main: str, pred: Callable[[Any], bool]) -> Predicate:
+    def _assess_sequence(*args, **kwargs) -> ExplanationOnFailure:
+        culprits = [(name, value) for name, value in join_args(args, kwargs) if not pred(value)]
+        if len(culprits) > 0:
+            return Explanation(expl_main, culprits)
+        return True
+    return _assess_sequence
+
+def approx(centre: Any, tol: Any = 1e-6, bounds: str = "[]") -> Predicate:
+    """
+    Returns a predicate satisfied when all arguments land within an interval defined by a centre and a tolerance
+    (which corresponds to its half-width). This can be used with any number-like objects that support addition and
+    some form of subtraction (either direct or by addition with opposite).
+
+    centre
+        Reference number against which we test for near-equality.
+    tol
+        Acceptable difference between arguments and the centre for the predicate to be satisfied.
+    bounds
+        Exclusivity of each implicit interval boundaries; see the docstring for `between` for further details.
+    """
+    upper = centre + tol
+    try:
+        lower = centre - tol
+    except TypeError:
+        try:
+            lower = centre + (-tol)
+        except TypeError:
+            raise TypeError("Operands do not support any form of subtraction")
+
+    return between(lower, upper, bounds=bounds)
+
+def belong_to(collection: Container) -> Predicate:
+    """
+    Predicate satisfied when all arguments are contained by the given collection, according to the `in'
+    operator.
+    """
+    return _assessor_sequence("These arguments did not belong to the collection", lambda v: v in collection)
+
+def is_any_of(*types: type) -> Predicate:
+    """
+    Predicate satisfied when all arguments are instances of one of the given types.
+    """
+    if len(types) < 1:
+        raise ValueError("Must provide at least one type.")
+    return _assessor_sequence(
+        "These arguments were not of type" + " or ".join(T.__name__ for T in types),
+        lambda v: any(isinstance(v, T) for T in types)
+    )
+
+def not_(predicate: Predicate) -> Predicate:
+    def opposite(*args: Any, **kwargs: Any) -> bool:
+        return not predicate(*args, **kwargs)
+    return opposite
+
 class Subscriber(ABC):
     """
     Object reacting to test results as they are generated by running tests.
     """
 
-    @abstractmethod
+    def on_add_to_suite(self, suite: "Suite") -> None:
+        pass
+
     def on_result(self, name_test: str, result: Result) -> None:
-        raise NotImplementedError()
+        pass
 
 TestFunction = Callable[..., None]
 
@@ -260,13 +512,9 @@ class Suite:
 
     def __init__(self, name_magic: str = "test") -> None:
         self._tests: Dict[str, List[Result]] = {}
-        self._fns_common = [fail, self.test, self._run_cell_test]
+        self._fns_common = [fail, self.test]
         self._subscribers: List[Subscriber] = []
 
-        ipython = get_ipython()
-        if ipython and name_magic:
-            register_cell_magic(name_magic)(lambda line, cell: self._test_cell(line, cell))
-
     def test(
         self,
         fn: Optional[TestFunction] = None,
@@ -318,9 +566,13 @@ class Suite:
             fn(*args, **kwargs)
             result = Success()
         except TestFailed as err:
-            result = Failure(err.reason or "Test marked as failed.", self._fns_common)
+            result = Failure(Explanation(err.reason, []), self._fns_common)
         except AssertionError as err:
-            result = Failure(str(err) or "Assertion failed.", self._fns_common)
+            if hasattr(err, "explanation"):
+                explanation = err.explanation
+            else:
+                explanation = Explanation(str(err), [])
+            result = Failure(explanation, self._fns_common)
         except BaseException:
             result = Error(self._fns_common)
 
@@ -339,30 +591,6 @@ class Suite:
 
         return fn
 
-    @staticmethod
-    def _run_cell_test(cell: str) -> None:
-        """
-        Helper function for running tests written into cells.
-        """
-        code_source = "\n" + cell
-        ipython = get_ipython()
-        name_cell = ipython.compile.cache(code_source)
-        code = compile(code_source, name_cell, "exec")
-        exec(code, ipython.user_global_ns, locals())
-
-    def _test_cell(self, line: str, cell: Optional[str]) -> None:
-        """
-        Runs a test written using a cell magic.
-        """
-        line = line.strip()
-        if not line:
-            raise ValueError("Please provide a title for the test (right after the cell magic invocation).")
-        cell = (cell or "").strip()
-        if not cell:
-            raise ValueError("There is no test to execute! Please write some code in there.")
-
-        self.test(self._run_cell_test, name=line, args=(cell,))
-
     @property
     def results(self) -> Iterator[Tuple[str, Iterator[Result]]]:
         """
@@ -389,6 +617,7 @@ class Suite:
         suite_with_subscriber._tests = copy(self._tests)  # Under assumption of results immutability.
         suite_with_subscriber._subscribers = copy(self._subscribers)
         suite_with_subscriber._subscribers.append(subscriber)
+        subscriber.on_add_to_suite(suite_with_subscriber)
         return suite_with_subscriber
 
 class Emphasis(ABC):
@@ -618,6 +847,9 @@ def summarize_results(  # noqa
         raise_on_error(suite)
     return summary
 
+def get_id_frame(frame: Frame) -> Sequence[str]:
+    return ([f"Function {frame.function}"] if frame.function else []) + [frame.name_file_pretty(True), f"Line {frame.num_line}"]
+
 def print_frame(  # noqa
     frame: Frame,
     file: TextIOBase = sys.stdout,
@@ -637,17 +869,7 @@ def print_frame(  # noqa
         Number of lines of code to fetch and write up before and after the
         line associated to the stack frame.
     """
-    header = (
-        colorizer.trivial
-        if Error.TAG_COMMON in frame.tags
-        else Plain()
-    )(
-        " | ".join([
-            "Code cell" if "cell" == frame.name_file else f"File {frame.name_file}",
-            f"Line {frame.num_line}",
-            f"Function {frame.function}"
-        ])
-    )
+    header = (colorizer.trivial if Error.TAG_COMMON in frame.tags else Plain())(" | ".join(get_id_frame(frame)))
     print(header, file=file)
     if Error.TAG_COMMON not in frame.tags:
         context: List[Tuple[int, str]] = frame.context(before=lines_context, after=lines_context)
@@ -668,7 +890,6 @@ def print_frame(  # noqa
                     sep=" ",
                     file=file
                 )
-    print(file=file)
 
 def detail_result(  # noqa
     name_test: str,
@@ -676,7 +897,8 @@ def detail_result(  # noqa
     prefix_header: str,
     file: TextIOBase = sys.stdout,
     colorizer: Colorizer = color(),
-    lines_context: int = 3
+    lines_context: int = 3,
+    print_border: bool = True
 ) -> None:
     """
     Writes up a report regarding a single test result.
@@ -694,22 +916,46 @@ def detail_result(  # noqa
     lines_context
         Number of lines of code to fetch and write up before and after the
         line associated to the stack frame.
+    print_border
+        If True (default), prints a border of dashes above the result report.
     """
     header = " ** ".join([
         prefix_header,
         f"Test {colorizer.important(name_test)}",
         {Failure: colorizer.failure, Error: colorizer.error}[type(result)](type(result).__name__)
     ])
-    print("-" * len(colors.strip_color(header)), file=file)
+    if print_border:
+        print("-" * len(colors.strip_color(header)), file=file)
     print(header, file=file)
-    if result.is_failure():
-        print(result.reason, file=file)
-    else:
-        print(f"{result.type_exc.__name__}:", str(result.value_exc) or "<no detail provided>", file=file)
-    print(file=file)
 
-    for frame in result.traceback:  # First frame is always Suite.test, which is irrelevant.
-        print_frame(frame, file=file, colorizer=colorizer, lines_context=lines_context)
+    if result.is_failure():
+        msg = result.explanation.main
+        if not msg:
+            if result.type_exc is TestFailed:
+                msg = "Test marked as failed"
+            elif result.type_exc is AssertionError:
+                msg = "Assertion failed"
+            else:
+                msg = "Unexpected failure (please take it up to the maintainer)"
+        if result.explanation.culprits:
+            width_column_name = min(16, max([len(str(n)) for n, v in result.explanation.culprits]))
+            msg += ":\n\n" + "\n".join(
+                f"    {name:{width_column_name}} => {value}"
+                for name, value in result.explanation.culprits
+            )
+        else:
+            msg += "."
+        print(msg, file=file)
+    else:
+        explanation = str(result.value_exc)
+        print(
+            f"{result.type_exc.__name__}" + (f": {explanation}" if explanation else " (no detail provided)"),
+            file=file,
+            end="\n\n"
+        )
+        for frame in result.traceback:  # First frame is always Suite.test, which is irrelevant.
+            print_frame(frame, file=file, colorizer=colorizer, lines_context=lines_context)
+            print(file=file)
 
 def detail_issues(  # noqa
     suite: Suite,
@@ -755,7 +1001,7 @@ def detail_issues(  # noqa
                 for r in rez:
                     if not r.is_success():
                         detail_result(name, r, f"# {index}/{num_problems}", lines_context=lines_context, file=file)
-                        print()
+                        print(file=file)
 
                         if index >= max_report:
                             num_remaining = num_problems - index
@@ -801,7 +1047,6 @@ class Report(Subscriber):
         lines_context: int = 3
     ) -> None:
         super().__init__()
-        self._last: Optional[bool] = None
         self._file = file
         self._verbose = verbose
         self._colorizer = colorizer
@@ -810,41 +1055,80 @@ class Report(Subscriber):
     def on_result(self, name_test: str, result: Result) -> None:
         name_test = self._colorizer.important(name_test)
         if result.is_success():
-            msg = f"Test {name_test} passed."
-            if self._verbose and self._last is False:
-                print("-" * len(colors.strip_color(msg)), file=self._file)
-            self._last = True
-            print(msg, file=self._file)
+            print(f"Test {name_test} passed.", file=self._file)
         else:
-            if self._last is True:
-                print(file=self._file)
-            self._last = False
             if self._verbose:
                 detail_result(
                     name_test,
                     result,
-                    "Issue encountered",
+                    "Issue",
                     self._file,
                     self._colorizer,
-                    self._lines_context
+                    self._lines_context,
+                    print_border=False
                 )
             else:
                 index_frame_relevant = -1
+                if result.type_exc in [TestFailed, DetailedAssertionError]:
+                    index_frame_relevant = -2
+                frame_relevant = result.traceback[index_frame_relevant]
+                details_frame = get_id_frame(frame_relevant)
                 if result.is_failure():
-                    label = self._colorizer.failure(f"Test {name_test} failed")
-                    print(f"{label}: {result.reason}", file=self._file)
-                    if isinstance(result.type_exc, TestFailed):
-                        index_frame_relevant = -2
+                    label = self._colorizer.failure(f"Failed test {name_test}")
+                    reason = result.reason
+                    if not reason:
+                        if result.type_exc is TestFailed:
+                            reason = "Deliberate failure trigger"
+                        elif result.type_exc is AssertionError:
+                            reason = "Unsatisfied assertion"
+                        else:
+                            reason = "Unexpected! report to maintainer"
+                    print(f"{label} -- {reason} | {' | '.join(details_frame)}", file=self._file)
                 else:
                     label = self._colorizer.error(f"Error occured during test {name_test}")
                     value_exc = ""
                     if str(result.value_exc):
                         value_exc = f" -- {str(result.value_exc)}"
                     print(f"{label}: {result.type_exc.__name__}{value_exc}", file=self._file)
-                frame_relevant = result.traceback[index_frame_relevant]
-                print_frame(
-                    frame_relevant,
-                    file=self._file,
-                    colorizer=self._colorizer,
-                    lines_context=self._lines_context
-                )
\ No newline at end of file
+                    print_frame(
+                        frame_relevant,
+                        file=self._file,
+                        colorizer=self._colorizer,
+                        lines_context=self._lines_context
+                    )
+
+class Magic(Subscriber):
+
+    def __init__(self, name_magic: str = "test") -> None:
+        self._name_magic = name_magic
+
+    def on_add_to_suite(self, suite):
+        suite._fns_common.append(run_test_from_cell)
+        ipython = get_ipython()
+        if ipython and self._name_magic:
+            register_cell_magic(self._name_magic)(lambda line, cell: test_cell(suite, line, cell))
+
+
+def test_cell(suite: Suite, line: str, cell: Optional[str]) -> None:
+    """
+    Runs a test written using a cell magic.
+    """
+    line = line.strip()
+    if not line:
+        raise ValueError("Please provide a title for the test (right after the cell magic invocation).")
+    cell = (cell or "").strip()
+    if not cell:
+        raise ValueError("There is no test to execute! Please write some code in there.")
+
+    suite.test(run_test_from_cell, name=line, args=(cell,))
+
+
+def run_test_from_cell(cell: str) -> None:
+    """
+    Executes the body of a cell, in context of the execution of a test.
+    """
+    code_source = "\n" + cell
+    ipython = get_ipython()
+    name_cell = ipython.compile.cache(code_source)
+    code = compile(code_source, name_cell, "exec")
+    exec(code, ipython.user_global_ns, locals())
\ No newline at end of file
